# Generated by Django 5.2.9 on 2025-12-15 11:44

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wbportfolio', '0094_alter_portfolio_depends_on_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='portfolio',
            name='only_keep_essential_positions',
            field=models.BooleanField(default=False, help_text='Only keep essential positions when drifting positions'),
        ),
    ]
