# Generated by Django 5.0.12 on 2025-11-06 12:59

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wbportfolio', '0090_dividendtransaction_price_fx_portfolio_and_more'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='order',
            name='execution_confirmed',
        ),
        migrations.AddField(
            model_name='order',
            name='execution_instruction',
            field=models.CharField(choices=[('MARKET_ON_CLOSE', 'Market On Close'), ('GUARANTEED_MARKET_ON_CLOSE', 'Guaranteed Market On Close'), ('GUARANTEED_MARKET_ON_OPEN', 'Guaranteed Market On Open'), ('GPW_MARKET_ON_CLOSE', 'GPW Market On Close'), ('MARKET_ON_OPEN', 'Market On Open'), ('IN_LINE_WITH_VOLUME', 'In Line With Volume'), ('LIMIT_ORDER', 'Limit Order'), ('VWAP', 'VWAP'), ('TWAP', 'TWAP')], default='MARKET_ON_CLOSE', max_length=26, verbose_name='Execution Instruction'),
        ),
        migrations.AddField(
            model_name='order',
            name='execution_instruction_parameters',
            field=models.JSONField(blank=True, default=dict, verbose_name='Execution Instruction Parameters'),
        ),
        migrations.AddField(
            model_name='order',
            name='execution_status',
            field=models.CharField(choices=[('PENDING', 'Pending'), ('CONFIRMED', 'Confirmed'), ('EXECUTED', 'Executed'), ('FAILED', 'Failed'), ('IGNORED', 'Ignored')], default='PENDING', max_length=12, verbose_name='Execution Status'),
        ),
    ]
