# Generated by Django 5.0.9 on 2024-11-11 11:59

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wbportfolio", "0070_remove_assetposition_unique_asset_position_and_more"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="trade",
            options={"ordering": ("order",), "verbose_name": "Trade", "verbose_name_plural": "Trades"},
        ),
        migrations.AlterField(
            model_name="trade",
            name="order",
            field=models.PositiveIntegerField(db_index=True, default=0, editable=False, verbose_name="order"),
            preserve_default=False,
        ),
    ]
