# Generated by Django 5.2.9 on 2026-02-04 08:50

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wbportfolio', '0096_portfoliotradableexchangethroughmodel'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='product',
            options={'permissions': [('administrate_product', 'Can administrate Product')], 'verbose_name': 'Product', 'verbose_name_plural': 'Products'},
        ),
        migrations.AddField(
            model_name='product',
            name='next_termination_reminder_date',
            field=models.DateField(blank=True, null=True),
        ),
    ]
