# Generated by Django 5.2.9 on 2025-12-09 08:53

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wbfdm', '0035_alter_instrument_related_instruments_and_more'),
        ('wbportfolio', '0093_remove_portfolioportfoliothroughmodel_unique_primary_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='portfolio',
            name='depends_on',
            field=models.ManyToManyField(blank=True, related_name='dependent_portfolios', through='wbportfolio.PortfolioPortfolioThroughModel', through_fields=('portfolio', 'dependency_portfolio'), to='wbportfolio.portfolio', verbose_name='The portfolios this portfolio depends on'),
        ),
        migrations.AlterField(
            model_name='portfolio',
            name='preferred_instrument_classifications',
            field=models.ManyToManyField(blank=True, limit_choices_to=models.Q(('instrument_type__is_classifiable', True), ('level', 0)), related_name='preferred_portfolio_classifications', through='wbportfolio.PortfolioInstrumentPreferredClassificationThroughModel', through_fields=('portfolio', 'instrument'), to='wbfdm.instrument', verbose_name='The Preferred classification per instrument'),
        ),
    ]
